/* 
	bblint.c	v0.13

	Checks a file for valid bB syntax, and issues warnings
	for anything that doesn't look right, along with the 
	line number in question if possible.

	This version checks for:

		-valid commands as labels.
		-indented "end" command.
		-unmatched for and next count.
		-incorrect "end" count.

	Checks that *may* be incorporated into future versions:

		-goto/gosub without real destination. (also searching includes)
		-dim a variable name that's a reserved keyword
		-setup  a constant that's a reserved keyword
		-invalid command

	Legal Stuff:

		No warranty of fitness is granted, nor shall the author he held
		liable for any damages resulting from the use of the program.

		The author reserves no copyright on the source code included
		with this package. You may use it as you see fit, including
		redistribution and/or relicensing the code or derivative code.

*/


#include "bblint.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

void err(char *errstring,...);
void warn(char *warnstring,...);
void check_badlabel(char *line, int linenumber);
void check_indentedend(char *line, int linenumber);	
void usage(void);
char *getcommandline(char *line);
char *getcommandword(char *line);
char *getlabel(char *line);
int check_endstart(char *line);
void getfield(char *line,int field);

const char *VERSION="v0.13";
#define LINEMAX 512

int main(int argc, char **argv)
{
	FILE *in;
	char line[LINEMAX];
	char rawline[LINEMAX];
	int linenumber;
	int t;
	int count_for, count_next;
	int count_endstart, count_endend;
        int last_endstart_line = -1;
        char last_endstart_text[LINEMAX];
	int field;

	if(argc==1)
		usage();
	in=fopen(argv[1],"r");
	if(in==NULL)
	{
		err("ERROR: couldn't open file \"%s\"\nExiting...\n",argv[1]);
	}
	
	linenumber=0;
	count_for=0;
	count_next=0;
	count_endstart=0;
	count_endend=0;
	while(fgets(rawline,LINEMAX,in)!=NULL) 
	{
		linenumber++;
		for(t=0;t<LINEMAX;t++)
		{
			//strip newlines...
			if ( (rawline[t]=='\n') ||
				(rawline[t]=='\r') )
				rawline[t]=0;

			//tabs to spaces...
			if ( rawline[t]=='\t')
				rawline[t]=' ';
		}

		rawline[LINEMAX-1]=0; //add a guard


		for(field=0;field<50;field++)
		{
			strcpy(line, rawline);

			getfield(line,field);

			//the line is empty
			if(line[0]==0)
				break;

			//a semicolon comment is valid, even without margin
			if(line[0]==';')
				break;

			//TODO: handle a /* comment with no margin

			//suck up any trailing spaces...
			for(t=0;t<LINEMAX;t++) 
			{
				if(line[t]==0)
				{
					t--;
					break;
				}
			}
			for(;t>0;t--)	
			{
				if(line[t]!=' ')
					break;
				line[t]=0;
			}

			//various start/end counts...
			if(check_endstart(line))
			{
				if((last_endstart_line>0)&&(strstr(last_endstart_text,"macro")==NULL))
					warn("warning(%d):\tmissing matching end to \"%s\"\n",last_endstart_line,last_endstart_text);
				count_endstart+=1;
				last_endstart_line=linenumber;
				strcpy(last_endstart_text,line);
			}
			//count_endstart=count_endstart+check_endstart(line);

			if(strcmp(getlabel(line),"end")==0)
			{
				count_endend++;
				last_endstart_line=-1;
			}
			if(strcmp(getcommandword(line),"for")==0)
			{
				count_for++;
			}
			if(strcmp(getcommandword(line),"next")==0)
			{
				count_next++;
			}
			check_badlabel(line,linenumber);
			check_indentedend(line,linenumber);	

			// We stop processing the line if we hit a comment
			if(strcmp(getcommandword(line),"rem")==0)
				break;
			if(*(getcommandword(line))==';')
				break;

			// TODO: handle a /* comment.
			// it should advance to the end */ marker within
			// the line. If there's no marker in the line, we
			// should keep reading stdin until we've found the
			// */ or have hit the end of the file.
		}
	}
	if(count_for!=count_next)
		warn("warning():\tFound %d \"for\" loops and %d \"next\" loop-endings.\n",count_for,count_next);
	if(count_endstart!=count_endend)
		warn("warning():\tFound %d commands that use end,  and %d \"end\" commands.\n",count_endstart,count_endend);
	fclose(in);
}

void check_badlabel(char *line,int linenumber)
{
	//check for labels that are named after keywords...

	int t;
	char label[LINEMAX];
	strncpy(label,getlabel(line),LINEMAX);
	label[LINEMAX-1]=0;

	for(t=0;labelblacklist[t]!=NULL;t++)
	{
		if(strncmp(labelblacklist[t],label,LINEMAX)==0)
		{
			warn("warning(%d):\tThe \"%s\" command appears to be missing indentation.\n",linenumber,label);
			return;
		}
	}
}


void check_indentedend(char *line,int linenumber)	
{
	//check for an indented word named "end"...
	char *endlocation;
	char *onebefore;
	int t;

	char command[LINEMAX];
	strncpy(command,getcommandword(line),LINEMAX);

	if(strncmp("end",command,LINEMAX)==0)
	{
		warn("warning(%d):\tThe \"end\" command appears to be indented.\n",linenumber,command);
	}
	return;
}

int check_endstart(char *line)
{
	// ** if command is one of those starting an "end" block, then return 1
	// ** otherwise return 0

	int t;
	char command[LINEMAX];

	strncpy(command,getcommandword(line),LINEMAX);
	command[LINEMAX-1]=0;

	if(command[0]==0)
		return(0);

	//return on rem, in case a keyword is in the comments
	if(strcmp("rem",command)==0)
		return(0);

	//some of the keywords show up in the set command 
	if(strcmp("set",command)==0)
		return(0);

	for(t=0;endcommands[t]!=NULL;t++)
	{
		if(strcmp(endcommands[t],command)==0)
		{
			return(1);
		}
	}

	return(0);
}

//the following get* functions would have been a *lot* simpler
//with regular expressions, but I wanted this to be ansi C...

void getfield(char *line,int fieldnumber)
{
	//fetches a field, defined by the : seperator, starting at 0
	//this takes the string in as input and modifies it for output
	char field[LINEMAX];
	char *fieldstart;
	char *fieldend;
	char *substitute;
	int fieldcount;
	int s,t;

	fieldcount=0;
	fieldstart=line;

	substitute=strstr(line," then ");
	if (substitute!=NULL)
	{
		substitute[1]=':';
		substitute[2]=' ';
		substitute[3]=' ';
		substitute[4]=' ';
	}
	substitute=strstr(line," else ");
	if (substitute!=NULL)
	{
		substitute[1]=':';
		substitute[2]=' ';
		substitute[3]=' ';
		substitute[4]=' ';
	}

	for(s=0;s<LINEMAX;s++)
	{
		if(line[s]==0)
			break;
		if(fieldcount==fieldnumber)
			break;
		if(s>5)
		{
			if((line[s]==':')&&(strncmp("lives:",line-5,6)!=0))
				fieldcount++;
		}
	}
	if(fieldcount<fieldnumber)
	{
		line[0]=0;
		return;
	}
	for(t=s;t<LINEMAX;t++)
	{
		if(line[t]==0)
			break;

		if((line[s]==':')&&(strncmp("lives:",line-5,6)!=0))
			break;
	}
	line[t]=0;
	if(fieldnumber>0)
	{
		s--;
		line[s]=' ';
	}
	strncpy(field,line+s,t-s);
	field[t-s]=0;
	
	strcpy(line,field);
	line[t]=0;
	return;
}

char *getlabel(char *line)
{
	// ** get label, if there is one...

	static char label[LINEMAX];
	int t;

	strncpy(label,line,LINEMAX);
	label[LINEMAX-1]=0;

	for(t=0;t<LINEMAX-1;t++)
	{
		if( (line[t]==0) ||
			(line[t]==' ') || (line[t]=='\t'))
		{
		label[t]=0;
		return(label);
		}
	}
	return(label);
}

char *getcommandline(char *line)
{
	// ** get the command-line after the label, if there is one
	int t;
	int commandstart;
	static char command[LINEMAX];

	command[0]=0;

	if(line[0]==0)
	{
		return(command);
	}
	t=0;
	if (line[0]!=' ')
	{
		//there's a label! skip it...
		for(t=0;t<LINEMAX-1;t++)
		{
			if(line[t]==0)
			{
				//just a label...
				return(command);
			}
			if(line[t]==' ')
			{
				break;
			}
		}
	}
	for(;t<LINEMAX;t++)
	{
	//now skip the spacing...
		if(line[t]==0)
		{
			//just a label...
			return(command);
		}
		if(line[t]!=' ')
		{
			break;
		}
	}
	commandstart=t;
	strcpy(command,line+commandstart);
	return(command);
}

char *getcommandword(char *line)
{
	// ** get the first word after the label, if there is one...
	int t;
	int commandstart, commandend;
	static char command[LINEMAX];

	command[0]=0;

	if(line[0]==0)
	{
		return(command);
	}
	t=0;
	if (line[0]!=' ')
	{
		//there's a label! skip it...
		for(t=0;t<LINEMAX-1;t++)
		{
			if(line[t]==0)
			{
				//just a label...
				return(command);
			}
			if(line[t]==' ')
			{
				break;
			}
		}
	}
	for(;t<LINEMAX;t++)
	{
	//now skip the spacing...
		if(line[t]==0)
		{
			//just a label...
			return(command);
		}
		if(line[t]!=' ')
		{
			break;
		}
	}
	commandstart=t;
	for(;t<LINEMAX;t++)
	{
	//now find the end of the command
		if ( 	(line[t]==0) ||
			(line[t]==' ') ||
			(line[t]==':')  )
		{
		commandend=t;
		break;
		}
	}

	if(strncmp("lives:",line+commandstart,6)==0) 
	{
	commandend++;
	}

	strncpy(command,line+commandstart,commandend-commandstart);
	command[commandend-commandstart]=0;
	return(command);
}

void err(char *errstring,...)
{
	va_list errargs;
	va_start(errargs, errstring);
	vprintf(errstring,errargs);
	va_end(errargs);
	exit(1);
}

void warn(char *warnstring,...)
{
	va_list warnargs;
	va_start(warnargs, warnstring);
	vprintf(warnstring,warnargs);
	va_end(warnargs);
}

void usage(void)
{
	printf("\nbblint %s\n\n",VERSION);
	printf("usage:\tbblint FILE\n\n");
	printf("\tChecks the specified basic file for possible syntax errors.\n\n");
	exit(0);
}
